/*
-------------------------------------------------------------------------------
    This file is a part of the VST Preset Generator
    Copyright (C) 2007  Francois Mazen

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
-------------------------------------------------------------------------------
  Name: vstpreset_test.cpp
  Author: Francois Mazen
  Date: 19/07/16 20:12
  Description: a test class for the VSTPreset class
-------------------------------------------------------------------------------
*/

#include "../../vstpreset.h"
#include "../../fxheader.h"
#include "../../parametertablemodel.h"
#include "../../vpgvsthost.h"
#include "../../endiannessconverter.h"

#include <QtTest>

#ifdef WIN64
    #define MOCKCHUNKPLUGIN_FILENAME "mockchunkplugin_win64.dll"
#elif WIN32
    #define MOCKCHUNKPLUGIN_FILENAME "mockchunkplugin_win32.dll"
#elif __x86_64
    #define MOCKCHUNKPLUGIN_FILENAME "libmockchunkplugin_linux.so"
#endif

#define STRINGIZE(x) #x
#define STRINGIZE_VALUE_OF(x) STRINGIZE(x)


class VSTPresetTest : public QObject
{
    Q_OBJECT

public:
    VSTPresetTest();

private:
    QFile readOnlyFile;
    QFileInfo testPresetFile;

    void removeWrittenFiles();
    void fillFxHeaderAndParametersForMockPlugin(FxHeader& fxHeader, ParameterTableModel& parameters);
    void fillFxHeaderAndParametersForMockChunkPlugin(FxHeader& fxHeader, ParameterTableModel& parameters, VpgVstHost& vstHost);
    void checkRawProgramFile(QFile &presetFile, const FxHeader& fxHeader);
    void checkRawChunkProgramFile(QFile &presetFile, const FxHeader& fxHeader);

private Q_SLOTS:

    void initTestCase();
    void cleanupTestCase();

    void cleanup();

    void testInvalidFilename();

    void testWriteProgramToFile();
    void testWriteBankToFile();

    void testWriteChunkProgramToFile();

    void testWriteChunkBankToFile();
};


VSTPresetTest::VSTPresetTest():
    testPresetFile("program_test_preset.fxp")
{
    this->testPresetFile.setCaching(false);
}

void VSTPresetTest::initTestCase()
{
    //Create a read only file.
    QString readOnlyFilename("readonlyfile");
    this->readOnlyFile.setFileName(readOnlyFilename);
    QVERIFY(this->readOnlyFile.open(QIODevice::WriteOnly));
    this->readOnlyFile.write("test");
    this->readOnlyFile.close();
    QVERIFY(this->readOnlyFile.setPermissions(QFileDevice::ReadOwner));

    removeWrittenFiles();
}

void VSTPresetTest::removeWrittenFiles()
{
    // Check that the files that will be generated don't exist.
    if(this->testPresetFile.exists())
    {
        QVERIFY(QFile::remove(this->testPresetFile.absoluteFilePath()));
    }
}

void VSTPresetTest::cleanupTestCase()
{
    // Remove read only file.
    QVERIFY(this->readOnlyFile.setPermissions(QFileDevice::WriteOwner));
    QVERIFY(this->readOnlyFile.remove());

    removeWrittenFiles();
}

void VSTPresetTest::cleanup()
{
    removeWrittenFiles();
}

void VSTPresetTest::fillFxHeaderAndParametersForMockPlugin(FxHeader &fxHeader, ParameterTableModel &parameters)
{
    // Create FxHeader and Parameters for the VPGMockPlugin.

    fxHeader.Data.fxID = "MCK1";
    fxHeader.Data.fxVersion = 1000;
    fxHeader.Data.nParams = 3;
    fxHeader.Data.useChunk = false;
    fxHeader.Data.name = "VPG Mock Plugin";

    parameters.beginModelChangeWithoutRefresh();
    parameters.clearWithoutRefresh();
    parameters.setParamNbWithoutRefresh(3);

    Parameter parameter(tr("unknown"));
    parameter.setRandomMode(Parameter::RandomMode_Constant);

    parameter.setConstantValue(1.0);
    parameters.replaceParameterWithoutRefresh(1, parameter);
    parameter.setConstantValue(0.5);
    parameters.replaceParameterWithoutRefresh(2, parameter);
    parameter.setConstantValue(0.0);
    parameters.replaceParameterWithoutRefresh(3, parameter);

    parameters.endModelChangeWithoutRefresh();
}

void VSTPresetTest::fillFxHeaderAndParametersForMockChunkPlugin(FxHeader &fxHeader, ParameterTableModel &parameters, VpgVstHost &vstHost)
{
    // Create FxHeader and Parameters for the VPGMockPlugin.

    fxHeader.Data.fxID = "MCK2";
    fxHeader.Data.fxVersion = 1000;
    fxHeader.Data.nParams = 3;
    fxHeader.Data.useChunk = true;
    fxHeader.Data.name = "VPG Mock Plugin";

    parameters.beginModelChangeWithoutRefresh();
    parameters.clearWithoutRefresh();
    parameters.setParamNbWithoutRefresh(3);

    Parameter parameter(tr("unknown"));
    parameter.setRandomMode(Parameter::RandomMode_Constant);

    parameter.setConstantValue(1.0);
    parameters.replaceParameterWithoutRefresh(1, parameter);
    parameter.setConstantValue(0.5);
    parameters.replaceParameterWithoutRefresh(2, parameter);
    parameter.setConstantValue(0.0);
    parameters.replaceParameterWithoutRefresh(3, parameter);

    parameters.endModelChangeWithoutRefresh();

    // Load the right plugin in the vst host.
    QDir mockPluginDir(STRINGIZE_VALUE_OF(MOCKPLUGINS_FOLDER));
    QFileInfo pluginFileInfo(mockPluginDir, MOCKCHUNKPLUGIN_FILENAME);
    QVERIFY(pluginFileInfo.exists());

    VpgVstHost::ERROR_CODE errorCode = vstHost.load(pluginFileInfo.absoluteFilePath());
    QCOMPARE(errorCode, VpgVstHost::ERROR_NO_ERROR);
    QCOMPARE(vstHost.getFxId(), CCONST('M','C','K','2'));
}

void VSTPresetTest::checkRawProgramFile(QFile &presetFile, const FxHeader &fxHeader)
{
    int32_t chunkMagic = EndiannessConverter::fromBigEndian(presetFile.read(4));
    QCOMPARE(chunkMagic, CCONST('C','c','n','K'));

    int32_t byteSize = EndiannessConverter::fromBigEndian(presetFile.read(4));
    QCOMPARE(byteSize, 60);

    int32_t fxMagic = EndiannessConverter::fromBigEndian(presetFile.read(4));
    QCOMPARE(fxMagic, CCONST('F','x','C','k'));

    int32_t version = EndiannessConverter::fromBigEndian(presetFile.read(4));
    QCOMPARE(version, 1);

    int32_t fxID = EndiannessConverter::fromBigEndian(presetFile.read(4));
    QCOMPARE(fxID, CCONST('M','C','K','1'));

    int32_t fxVersion = EndiannessConverter::fromBigEndian(presetFile.read(4));
    QCOMPARE(fxVersion, fxHeader.Data.fxVersion);

    int32_t numParams = EndiannessConverter::fromBigEndian(presetFile.read(4));
    QCOMPARE(numParams, fxHeader.Data.nParams);

    QString prgName(presetFile.read(28).data());
    QVERIFY(prgName.startsWith("VPG"));

    int32_t param1IntValue = EndiannessConverter::fromBigEndian(presetFile.read(4));
    QCOMPARE(1.0, *(float*)(&param1IntValue));

    int32_t param2IntValue = EndiannessConverter::fromBigEndian(presetFile.read(4));
    QCOMPARE(0.5, *(float*)(&param2IntValue));

    int32_t param3IntValue = EndiannessConverter::fromBigEndian(presetFile.read(4));
    QCOMPARE(0.0, *(float*)(&param3IntValue));
}

void VSTPresetTest::checkRawChunkProgramFile(QFile &presetFile, const FxHeader &fxHeader)
{
    QString referenceChunk = "PROGRAM;1;0.5;0";

    int32_t chunkMagic = EndiannessConverter::fromBigEndian(presetFile.read(4));
    QCOMPARE(chunkMagic, CCONST('C','c','n','K'));

    int32_t referenceSize = 4*5 + 28 + 4 + referenceChunk.size();
    int32_t byteSize = EndiannessConverter::fromBigEndian(presetFile.read(4));
    QCOMPARE(byteSize, referenceSize);

    int32_t fxMagic = EndiannessConverter::fromBigEndian(presetFile.read(4));
    QCOMPARE(fxMagic, CCONST('F','P','C','h'));

    int32_t version = EndiannessConverter::fromBigEndian(presetFile.read(4));
    QCOMPARE(version, 1);

    int32_t fxID = EndiannessConverter::fromBigEndian(presetFile.read(4));
    QCOMPARE(fxID, CCONST('M','C','K','2'));

    int32_t fxVersion = EndiannessConverter::fromBigEndian(presetFile.read(4));
    QCOMPARE(fxVersion, fxHeader.Data.fxVersion);

    int32_t numParams = EndiannessConverter::fromBigEndian(presetFile.read(4));
    QCOMPARE(numParams, fxHeader.Data.nParams);

    QString prgName(presetFile.read(28).data());
    QVERIFY(prgName.startsWith("VPG"));

    int32_t chunkSize = EndiannessConverter::fromBigEndian(presetFile.read(4));
    QCOMPARE(chunkSize, referenceChunk.size());

    QByteArray chunk = presetFile.read(referenceChunk.size());
    QCOMPARE(chunk.size(), referenceChunk.size());
    QCOMPARE(QString(chunk), referenceChunk);
}

void VSTPresetTest::testInvalidFilename()
{
    QFileInfo fileInfo(this->readOnlyFile);

    FxHeader dummyFxHeader;
    ParameterTableModel dummyParameters;
    VpgVstHost dummyHost;

    QCOMPARE(VSTPreset::WriteProgramToFile(fileInfo.absoluteFilePath(), dummyFxHeader, dummyParameters), VSTPreset::ERROR_COULD_NOT_CREATE_FILE);
    QCOMPARE(VSTPreset::WriteProgramChunkToFile(fileInfo.absoluteFilePath(), dummyFxHeader, dummyParameters, dummyHost), VSTPreset::ERROR_COULD_NOT_CREATE_FILE);
    QCOMPARE(VSTPreset::WriteBankToFile(fileInfo.absoluteFilePath(), dummyFxHeader, dummyParameters, 10), VSTPreset::ERROR_COULD_NOT_CREATE_FILE);
    QCOMPARE(VSTPreset::WriteBankChunkToFile(fileInfo.absoluteFilePath(), dummyFxHeader, dummyParameters, dummyHost, 10), VSTPreset::ERROR_COULD_NOT_CREATE_FILE);
}

void VSTPresetTest::testWriteProgramToFile()
{
    FxHeader fxHeader;
    ParameterTableModel parameters;
    fillFxHeaderAndParametersForMockPlugin(fxHeader, parameters);

    // Write the preset file.
    QCOMPARE(this->testPresetFile.exists(), false);

    VSTPreset::ERROR_CODE errorCode = VSTPreset::WriteProgramToFile(this->testPresetFile.absoluteFilePath(), fxHeader, parameters);

    QCOMPARE(errorCode, VSTPreset::ERROR_NO_ERROR);
    QCOMPARE(this->testPresetFile.exists(), true);

    // Check preset file content.
    QFile presetFile(this->testPresetFile.absoluteFilePath());
    QVERIFY(presetFile.open(QIODevice::ReadOnly));

    checkRawProgramFile(presetFile, fxHeader);

    presetFile.close();
}

void VSTPresetTest::testWriteBankToFile()
{
    FxHeader fxHeader;
    ParameterTableModel parameters;
    fillFxHeaderAndParametersForMockPlugin(fxHeader, parameters);

    // Write the preset file.
    QCOMPARE(this->testPresetFile.exists(), false);

    int programCountPerBank = 5;
    VSTPreset::ERROR_CODE errorCode = VSTPreset::WriteBankToFile(this->testPresetFile.absoluteFilePath(), fxHeader, parameters, programCountPerBank);

    QCOMPARE(errorCode, VSTPreset::ERROR_NO_ERROR);
    QCOMPARE(this->testPresetFile.exists(), true);

    // Check preset file content.
    QFile presetFile(this->testPresetFile.absoluteFilePath());
    QVERIFY(presetFile.open(QIODevice::ReadOnly));

    int32_t chunkMagic = EndiannessConverter::fromBigEndian(presetFile.read(4));
    QCOMPARE(chunkMagic, CCONST('C','c','n','K'));

    int32_t byteSize = EndiannessConverter::fromBigEndian(presetFile.read(4));
    QCOMPARE(byteSize, 5*4+128+68*programCountPerBank); // 448

    int32_t fxMagic = EndiannessConverter::fromBigEndian(presetFile.read(4));
    QCOMPARE(fxMagic, CCONST('F','x','B','k'));

    int32_t version = EndiannessConverter::fromBigEndian(presetFile.read(4));
    QCOMPARE(version, 1);

    int32_t fxID = EndiannessConverter::fromBigEndian(presetFile.read(4));
    QCOMPARE(fxID, CCONST('M','C','K','1'));

    int32_t fxVersion = EndiannessConverter::fromBigEndian(presetFile.read(4));
    QCOMPARE(fxVersion, fxHeader.Data.fxVersion);

    int32_t numPrograms = EndiannessConverter::fromBigEndian(presetFile.read(4));
    QCOMPARE(numPrograms, programCountPerBank);

    int32_t currentProgram = EndiannessConverter::fromBigEndian(presetFile.read(4));
    QCOMPARE(currentProgram, 1);

    // Ignore next 124 bytes (reserved for future use).
    presetFile.read(124);

    // Check each program.
    for(int programIndex = 0; programIndex < programCountPerBank; programIndex++)
    {
        checkRawProgramFile(presetFile, fxHeader);
    }

    presetFile.close();
}

void VSTPresetTest::testWriteChunkProgramToFile()
{
    FxHeader fxHeader;
    ParameterTableModel parameters;
    VpgVstHost vstHost;
    fillFxHeaderAndParametersForMockChunkPlugin(fxHeader, parameters, vstHost);

    // Write the preset file.
    QCOMPARE(this->testPresetFile.exists(), false);

    VSTPreset::ERROR_CODE errorCode = VSTPreset::WriteProgramChunkToFile(this->testPresetFile.absoluteFilePath(), fxHeader, parameters, vstHost);

    QCOMPARE(errorCode, VSTPreset::ERROR_NO_ERROR);
    QCOMPARE(this->testPresetFile.exists(), true);

    // Check preset file content.
    QFile presetFile(this->testPresetFile.absoluteFilePath());
    QVERIFY(presetFile.open(QIODevice::ReadOnly));

    checkRawChunkProgramFile(presetFile, fxHeader);

    presetFile.close();

    vstHost.close();
}

void VSTPresetTest::testWriteChunkBankToFile()
{
    FxHeader fxHeader;
    ParameterTableModel parameters;
    VpgVstHost vstHost;
    fillFxHeaderAndParametersForMockChunkPlugin(fxHeader, parameters, vstHost);

    // Write the preset file.
    QCOMPARE(this->testPresetFile.exists(), false);

    int programCountPerBank = 5;
    VSTPreset::ERROR_CODE errorCode = VSTPreset::WriteBankChunkToFile(this->testPresetFile.absoluteFilePath(), fxHeader, parameters, vstHost, programCountPerBank);

    QCOMPARE(errorCode, VSTPreset::ERROR_NO_ERROR);
    QCOMPARE(this->testPresetFile.exists(), true);

    // Check preset file content.
    QFile presetFile(this->testPresetFile.absoluteFilePath());
    QVERIFY(presetFile.open(QIODevice::ReadOnly));

    int32_t chunkMagic = EndiannessConverter::fromBigEndian(presetFile.read(4));
    QCOMPARE(chunkMagic, CCONST('C','c','n','K'));

    QString referenceChunk = "BANK;1;0.5;0";
    int32_t byteSize = EndiannessConverter::fromBigEndian(presetFile.read(4));
    QCOMPARE(byteSize, 5*4 + 128 + 4 + referenceChunk.size());

    int32_t fxMagic = EndiannessConverter::fromBigEndian(presetFile.read(4));
    QCOMPARE(fxMagic, CCONST('F','B','C','h'));

    int32_t version = EndiannessConverter::fromBigEndian(presetFile.read(4));
    QCOMPARE(version, 1);

    int32_t fxID = EndiannessConverter::fromBigEndian(presetFile.read(4));
    QCOMPARE(fxID, CCONST('M','C','K','2'));

    int32_t fxVersion = EndiannessConverter::fromBigEndian(presetFile.read(4));
    QCOMPARE(fxVersion, fxHeader.Data.fxVersion);

    int32_t numPrograms = EndiannessConverter::fromBigEndian(presetFile.read(4));
    QCOMPARE(numPrograms, std::min(programCountPerBank, vstHost.getNumberPrograms()));

    int32_t currentProgram = EndiannessConverter::fromBigEndian(presetFile.read(4));
    QCOMPARE(currentProgram, 1);

    // Ignore next 124 bytes (reserved for future use).
    presetFile.read(124);

    // Check program chunk.
    int32_t chunkSize = EndiannessConverter::fromBigEndian(presetFile.read(4));
    QCOMPARE(chunkSize, referenceChunk.size());

    QByteArray chunk = presetFile.read(referenceChunk.size());
    QCOMPARE(chunk.size(), referenceChunk.size());
    QCOMPARE(QString(chunk), referenceChunk);

    presetFile.close();

}

QTEST_GUILESS_MAIN(VSTPresetTest)

#include "vstpreset_test.moc"
